
function compute_all_targets!(prob::Problem,
                              comm::MPI.Comm;
                             )
    rank = MPI.Comm_rank(comm)
    if rank == 0
        target_values = compute_all_targets(prob, comm)
        foreach(p -> prob.targets[p.first].current_value = p.second, target_values)
    else
        compute_all_targets(prob, comm)
    end
end

function compute_all_targets(prob::Problem,
                             comm::MPI.Comm;
                            )
    return compute_targets(prob.targets, prob.equilibrium_wrappers, comm)
end

function compute_targets(targets::Dict{Int, Target},
                         equilibrium_wrappers::Union{EquilibriumWrapper, Dict{UInt, EquilibriumWrapper}},
                         comm::MPI.Comm;
                        )
    rank = MPI.Comm_rank(comm)
    if rank == 0
        T = typeof(first(targets).second.target_value)
        target_values =Dict{Int, T}()
        wrappers = equilibrium_wrappers isa Dict ? equilibrium_wrappers : Dict(one(UInt) => equilibrium_wrappers)
        for eq_wrapper in values(wrappers)
            for key in keys(eq_wrapper.derived_geometries)
                MPI.bcast(false, 0, comm)
                MPI.bcast(key, 0, comm)
                geometry = extract_derived_geometry(eq_wrapper, key)
                MPI.bcast(geometry, 0, comm)
                geometry_targets = filter(t -> t.second.geometry_key == key, targets)
                MPI.bcast(geometry_targets, 0, comm)
                @debug "Rank $rank is calling targets with geometry $key"
                results = compute_targets(deepcopy(geometry_targets), geometry, comm)
                foreach(p -> target_values[p.first] = p.second, results)
            end
        end
        MPI.bcast(true, 0, comm)
        return target_values
    else
        finished = false
        while !finished
            finished = MPI.bcast(nothing, 0, comm)
            if !finished
                key = MPI.bcast(nothing, 0, comm)
                geometry = MPI.bcast(nothing, 0, comm)
                geometry_targets = MPI.bcast(nothing, 0, comm)
                @debug "Rank $rank is calling targets with geometry $key"
                compute_targets(geometry_targets, geometry, comm)
            end
        end
    end
end
        

function compute_targets(targets::Dict{Int, Target},
                         geometry::E,
                         comm::MPI.Comm;
                        ) where {E <: AbstractGeometry}
    rank = MPI.Comm_rank(comm)
    if rank == 0
        T = typeof(first(targets).second.target_value)
        target_values = Dict{Int, T}()
        for (i, target) in targets
            MPI.bcast(false, 0, comm)
            MPI.bcast(target, 0, comm)
            @debug "Rank $rank is calling $(target.label)"
            target_values[i] = compute_target(target, geometry, comm)
        end
        MPI.bcast(true, 0, comm)
        return target_values
    else
        finished = false
        while !finished
            finished = MPI.bcast(nothing, 0, comm)
            if !finished
                target = MPI.bcast(nothing, 0, comm)
                @debug "Rank $rank is calling $(target.label)"
                compute_target(target, geometry, comm)
            end
        end
    end
end

function compute_target(target::Target,
                        geometry::E,
                        comm::MPI.Comm;
                       ) where {T, E <: AbstractGeometry}
    rank = MPI.Comm_rank(comm)
    if rank == 0
        has_comm = :mpi_comm in target.method.args
        MPI.bcast(has_comm, 0, comm)
        target_method = Expr(target.method.head, 
                             replace(target.method.args, 
                                     target.geometry_key => geometry,
                                     :mpi_comm => comm)...)
        MPI.bcast(target_method, 0, comm)
        @debug "Rank $rank is calling $(target_method.args[1])"
        return eval(target_method)
    else
        has_comm = MPI.bcast(nothing, 0, comm)
        target_method = MPI.bcast(nothing, 0, comm)
        if has_comm
            @debug "Rank $rank is calling $(target_method.args[1])"
            eval(target_method)
        end
    end
end
